/****************************************************************
BUF_GAP.C -- Buffer gap manager

Derived by Ray Valdes from code by Joseph H. Allen, who posted
this to the comp.editors newsgroup on 9/10/89, and who wrote:
"Do whatever you like with this, just leave my name on it."
****************************************************************/

#include <stdio.h>

#include "config.h"
#include "std_defs.h"


/*------------------ Macros ------------------------*/

#define vb_SizeofGap() (theEndOfGap - theStartOfGap)

#define vb_CharAtPoint() \
    (thePoint==theStartOfGap \
	? *(thePoint=theEndOfGap)\
	: *thePoint)

#define vb_ReplaceCharAtPoint(c) \
    (((thePoint==theStartOfGap)? thePoint=theEndOfGap : 0), \
    ((thePoint==theEndOfBuffer) \
	? (vb_ExpandBuffer(1),theEndOfBuffer++) \
	: 0),\
      *thePoint=(c),\
      isBufferChanged=TRUE)

#define vb_GetCharMovePoint() \
    ((thePoint==theStartOfGap)\
	? (thePoint=theEndOfGap+1,*theEndOfGap)\
	: *(thePoint++))

#define vb_PutCharMovePoint(c) \
    (((thePoint==theStartOfGap)?thePoint=theEndOfGap:0),\
    ((thePoint==theEndOfBuffer)?(vb_ExpandBuffer(1),theEndOfBuffer++):0),\
    *(thePoint++) = (c),\
    isBufferChanged=TRUE)

#define vb_InsertCharAtPoint(c) \
    ((thePoint!=theStartOfGap?vb_MoveGapToPoint():0),\
    (theStartOfGap==theEndOfGap?vb_ExpandGap(1):0), \
     *(theStartOfGap++) = (c), \
     isBufferChanged=TRUE)

#define vb_DeleteChars(x) \
    ((thePoint!=theStartOfGap?vb_MoveGapToPoint():0), \
     theEndOfGap += (x), \
     isBufferChanged=TRUE)

#define vb_NumStartToPoint() \
    ((thePoint>=theEndOfGap) \
      ? (thePoint - theBuffer)-(theEndOfGap-theStartOfGap)\
      : thePoint - theBuffer)

#define vb_NumTotalChars() \
    ((theEndOfBuffer - theBuffer)-(theEndOfGap-theStartOfGap))

#define vb_AtEndOfFile() \
    ((thePoint==theStartOfGap)\
     ? (theEndOfGap==theEndOfBuffer)\
     : (thePoint==theEndOfBuffer))

#define vb_SetPoint(x) \
    (thePoint=theBuffer+(x), \
    (thePoint > theStartOfGap)\
	? (thePoint += theEndOfGap-theStartOfGap)\
	: 0)

#define vb_BackOneGetChar() \
    ((thePoint==theEndOfGap)\
	? *(thePoint=theStartOfGap-1)\
	: *(--thePoint))

#define vb_FowOneFindNewline() \
    (vb_AtEndOfFile() \
      ? 0              \
      : (vb_GetCharMovePoint(),vb_FindNewline()))

#define vb_BackOnePrevNewline() \
    (vb_NumStartToPoint() \
     ? (vb_BackOneGetChar(),vb_PrevNewline())\
     : 0)

/****************************************************************/

PUBLIC VOID 	vb_MoveBytes	(char *dst,char *src, unsigned s);
PUBLIC BOOL 	vb_InitVBuff	(VOID);    
PUBLIC VOID 	vb_FiniVBuff	(VOID);    

PUBLIC VOID 	vb_ExpandBuffer	(unsigned amount);
PUBLIC VOID 	vb_MoveGapToPoint(VOID);
PUBLIC VOID 	vb_ExpandGap	(unsigned size);
PUBLIC BOOL 	vb_FindNewline	(VOID);
PUBLIC BOOL 	vb_PrevNewline	(VOID);
PUBLIC VOID 	vb_InsertString	(char *string,    unsigned size);
PUBLIC BOOL 	vb_CompareString(char *string,   unsigned size);
PUBLIC int 	vb_ToUpper	(int c);
PUBLIC BOOL 	vb_CompareNoCase(char *string,  unsigned size);
PUBLIC BOOL 	vb_SaveFile	(FILE *file,  unsigned size);
PUBLIC int 	vb_InsertFile	(FILE* file);  /*assumes file is opened*/

/****************************************************************/

PRIVATE unsigned sizeofBuffer;	/* Size of theBuffer */
PRIVATE char*  thePoint;	/* The point */
PRIVATE char*  theBuffer;	/* The buffer */
PRIVATE char*  theEndOfBuffer;	/* First character not in buffer */
PRIVATE char*  theStartOfGap;	/* Beginning of theStartOfGap */
PRIVATE char*  theEndOfGap;	/* First character not in theStartOfGap */
PRIVATE BOOL   isBufferChanged;	/* Set when file has been changed */

#define SIZEOF_GAP_INCREMENT   16384  /* Amt that the buffer gap grows by */


/*---------------------- FUNCTIONS ---------------------*/

PUBLIC VOID 
vb_MoveBytes(
        char *dst,
        char *src,
        unsigned s)
{
    if(src==dst || s==0) return;

    if(src > dst)
    {
	do { *(dst++)= *(src++); } while(--s);
    }
    else
    {
	dst += s;
	src += s;
	do { *(--dst)= *(--src); } while(--s);
    }
}

/****************************************************************/    
PUBLIC VOID
vb_FiniVBuff(VOID)
{
    if(theBuffer) mem_FreeMem(theBuffer);
}

/****************************************************************/    
PUBLIC BOOL 
vb_InitVBuff(VOID)
{
    sizeofBuffer = SIZEOF_GAP_INCREMENT;
    theBuffer = (char *) mem_AllocMem(sizeofBuffer);
    if(!theBuffer) return FALSE;
    thePoint      = theBuffer;
    theStartOfGap = theBuffer;
    theEndOfGap   = theBuffer + SIZEOF_GAP_INCREMENT;
    theEndOfBuffer= theEndOfGap;
    return TRUE;
}

/****************************************************************/    
PUBLIC VOID 
vb_ExpandBuffer(unsigned amount)
{
    if( (theEndOfBuffer + amount - theBuffer) > sizeofBuffer)
    {
	char *old = theBuffer;
	sizeofBuffer = theEndOfBuffer +amount +SIZEOF_GAP_INCREMENT-theBuffer;
	theBuffer = (char *) realloc(theBuffer, sizeofBuffer);
	if(!theBuffer) ProgramError("realloc failed!");
	thePoint       += theBuffer - old;
	theEndOfBuffer += theBuffer - old;
	theStartOfGap  += theBuffer - old;
	theEndOfGap    += theBuffer - old;
    }
}

/****************************************************************/    
PUBLIC VOID 
vb_MoveGapToPoint(VOID)
{
    if(thePoint==theStartOfGap) return;

    if(thePoint==theEndOfGap)
    {
	thePoint = theStartOfGap;
	return;
    }
    /*else*/
    if(thePoint < theStartOfGap)
    {
	vb_MoveBytes( theEndOfGap - (theStartOfGap-thePoint),
			thePoint,
			theStartOfGap - thePoint);
        theEndOfGap   = theEndOfGap-(theStartOfGap-thePoint);
	theStartOfGap = thePoint;
    }
    else
    {
        vb_MoveBytes(theStartOfGap,theEndOfGap,thePoint-theEndOfGap);
        theStartOfGap += thePoint-theEndOfGap;
        theEndOfGap    = thePoint;
        thePoint       = theStartOfGap;
    }
}

/****************************************************************/    
PUBLIC VOID
vb_ExpandGap(unsigned size)
{
    if(size > vb_SizeofGap())
    {
	size += SIZEOF_GAP_INCREMENT;
	vb_ExpandBuffer(size);
	vb_MoveBytes(theEndOfGap+size,
			theEndOfGap, 
			theEndOfBuffer - theEndOfGap);
	theEndOfGap    += size;
	theEndOfBuffer += size;
    }
}

/****************************************************************/    
PUBLIC BOOL 
vb_FindNewline(VOID)
{
    while(((thePoint==theStartOfGap)
	    ? (thePoint=theEndOfGap)
	    : thePoint)
	  != theEndOfBuffer)
    {
	if(*thePoint=='\n') return TRUE;
	else thePoint++;
    }
    return FALSE;
}

/****************************************************************/    
PUBLIC BOOL 
vb_PrevNewline(VOID)
{
    if(vb_CharAtPoint()=='\n') return TRUE;

    while((thePoint==theEndOfGap
	? thePoint=theStartOfGap
	: thePoint)
	!=theBuffer)
     {
	 if( *(--thePoint) == '\n') return TRUE;
     }
    return FALSE;
}

/****************************************************************/    
PUBLIC VOID 
vb_InsertString(
    char *string,
    unsigned size)
{
    vb_MoveGapToPoint();
    if(size > vb_SizeofGap())	vb_ExpandGap(size);
    vb_MoveBytes(theStartOfGap,string,size);
    theStartOfGap += size;
    isBufferChanged = TRUE;
}

/****************************************************************/    
PUBLIC BOOL 
vb_CompareString(
    char *string,
    unsigned size)
{
    char *x;

    if(thePoint==theStartOfGap) thePoint=theEndOfGap;

    if(    theStartOfGap > thePoint 
	&& theStartOfGap < thePoint + size 
	&& theStartOfGap != theEndOfGap)
    {
     if(vb_CompareString(string,theStartOfGap-thePoint)) return TRUE;
     else
     {
	 x = thePoint;
	 thePoint = theEndOfGap;
	 if(vb_CompareString( string + (theStartOfGap-x),
				 size - (theStartOfGap-x)))
         { thePoint=x;	  return TRUE;   }
	 else
         {  thePoint=x;	  return FALSE;  }
     }
    }
    else
    {
	x = thePoint;
	do
	{ if(*(x++)!=*(string++)) return TRUE; }
	while(--size);
	return FALSE;
    }
}

/****************************************************************/    
PUBLIC int 
vb_ToUpper(int c)
{
    if(c >= 'a' && c <= 'z') return (c + 'A' - 'a');
    else return c;
}
    
/****************************************************************/    
PUBLIC BOOL
vb_CompareNoCase(
    char *string,
    unsigned size)
{
    char *x;

    if(thePoint==theStartOfGap) thePoint=theEndOfGap;

    if(    theStartOfGap >  thePoint 
	&& theStartOfGap <  thePoint+size 
	&& theStartOfGap != theEndOfGap)
    {

     if(vb_CompareString(string,theStartOfGap-thePoint)) 
	 return TRUE;
     else
     {
	 x = thePoint;
	 thePoint = theEndOfGap;
	 if(vb_CompareString( string + (theStartOfGap-x),
				 size - (theStartOfGap-x)))
	 { thePoint=x;   return TRUE;  }
	 else
	 { thePoint=x;   return FALSE; }
      }
    }
    else
    {
	x = thePoint;
	do
        {
	    if(vb_ToUpper(*(x++)) != vb_ToUpper(*(string++))) 
		return TRUE;
         }
	 while(--size);
	 return FALSE;
     }
}
    
/****************************************************************/    
PUBLIC BOOL 
vb_SaveFile(
    FILE *file,
    unsigned size) /*assumes file has been opened*/
{
    if(!size) return TRUE;

    if(thePoint==theStartOfGap) thePoint = theEndOfGap;

    if(    theStartOfGap >  thePoint 
	&& theStartOfGap <  thePoint+size 
	&& theStartOfGap != theEndOfGap)
   {
       if(theStartOfGap-thePoint !=
	 fwrite(thePoint,1,theStartOfGap-thePoint,file)) 
	 return FALSE;
       if(size - (theStartOfGap-thePoint)!=
	 fwrite(theEndOfGap,1,size-(theStartOfGap-thePoint),file)) 
	 return FALSE;
       return TRUE;
    }
    else
     return size==fwrite(thePoint,1,size,file);
}

/****************************************************************/

PUBLIC BOOL
vb_ReadFile(char *file_name)
{
    FILE *input_file;
    
    input_file = fopen(file_name,"rb");
    if(input_file==NULL)
    {
	FatalError(fmtStr("Can't open file %s",file_name));
    }
    dbg_PutLine(fmtStr("File %s is open.",file_name));
    if(!vb_InsertFile(input_file)) 
    {
	fclose(input_file);
	return FALSE;
    }
    /*else*/
    fclose(input_file);
    return TRUE;
}

/****************************************************************/    
#ifdef TURBO_C
#include <io.h>
#endif

PUBLIC int 
vb_InsertFile(FILE* file)  /*assumes file has been opened*/
{
    unsigned amount;
    long file_size;
    
#ifdef TURBO_C
    file_size = filelength(fileno(file));
#else
    struct stat buf;
    fstat(fileno(file),&buf);
    file_size = buf.st_size;
#endif

    if(file_size==0) return TRUE;
    if(file_size > 32767) return FALSE;
    
    isBufferChanged = TRUE;

    vb_MoveGapToPoint();
    vb_ExpandGap((int)file_size);

    amount = fread(theStartOfGap, 1, file_size, file);
    theStartOfGap += amount;

    return (amount==file_size);
}

