/****************************************************************
EDITLINE.C -- The simplest example of text editing routine
****************************************************************/

#include "std_defs.h"
#include "config.h"
#include "globals.h"

#include "keyboard.h"


void EditLine(char* buffer, int max_length, int curr_row)
{
    int c, str_length  = strlen(buffer), 
    	   curr_column = str_length, 
    	   insert_mode = TRUE;

    /*ChangeCursorShape(insert_mode);*/
    vt_ClrLinAt(curr_row,0);

    do
    {
	vt_StringAt(buffer, curr_row,0);
	vt_CursorAt(curr_row,curr_column);

	switch(c = vt_GetKeystroke())	/* dispatch on keystroke */
	{
	     /*-------- keystrokes that terminate the editing session-----*/
	     case UP_KEY:   	 			/* fall through */
	     case DOWN_KEY: 	 c = ENTER_KEY;		/* fall through */
	     case ENTER_KEY:     break;

	     case ESCAPE_KEY:    str_length = 0;                 break;

	    /*--------- keystrokes that merely change the cursor position---*/

	    case HOME_KEY:      curr_column = 0;     	  	 break;

	    case END_KEY:       curr_column = str_length;        break;

	    case INS_KEY:       insert_mode = !insert_mode;
				/*ChangeCursorShape(insert_mode);*/
				break;

            case LEFT_KEY:      if (curr_column > 0) curr_column--;
				break;

	    case RIGHT_KEY:     if (curr_column < str_length) curr_column++;
				break;

	    /*------ keystrokes that alter the contents of the buffer----*/

	    case BACKSPACE_KEY: if (curr_column > 0)
				{
				    movmem( &buffer[curr_column], 
					    &buffer[curr_column - 1], 
					    str_length - curr_column + 1);
				    curr_column--;  str_length--;
				}
				break;

	     case DELETE_KEY:   if (curr_column < str_length)
				{
				    movmem( &buffer[curr_column + 1], 
					    &buffer[curr_column], 
					    str_length - curr_column);
				    str_length--;
				 }
				 break;

	     default:		 if (((c >= ' ') && (c <= '~')) &&
	     				(str_length < max_length))
				 {
				 	if (insert_mode)
					{
					    movmem( &buffer[curr_column + 1], 
					    	&buffer[curr_column], 
						str_length - curr_column + 1);
					    str_length++;
					}
					else if (curr_column >= str_length)
					    str_length++;

					buffer[curr_column++] = c;
				  }
				  break;
	     }
	     buffer[str_length] = '\0';
    }
    while ((c != ENTER_KEY) && (c != ESCAPE_KEY));

    /*ClearInputLine();*/
    /*ClearCursor();*/
}

